/*
 * datamodel.c - XMPP datamodel plugin
 *
 * Copyright (C) 2022, IOPSYS Software Solutions AB.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * See LICENSE file for license related information.
 *
 */

#include <libbbfdm-api/dmbbf.h>
#include <libbbfdm-api/dmcommon.h>
#include <libbbfdm-api/dmuci.h>

#include "datamodel.h"

#define XMPP_STATS_PATH	"/tmp/.xmppc_stats"

typedef struct xmpp_connection_data {
	struct dmmap_dup *p;
	char rx_msg[32];
	char rx_err_msg[32];
	char tx_msg[32];
	char tx_err_msg[32];
} connection_data;

/* ********** DynamicObj ********** */
DM_MAP_OBJ tDynamicObj[] = {
/* parentobj, nextobject, parameter */
{"Device.", tDeviceXMPPObj, NULL},
{"Device.ManagementServer.", NULL, tDeviceManagementServerParam},
{0}
};

/**************************************************************************
* LINKER
***************************************************************************/
static int get_linker_xmpp_connection(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	*linker = dmstrdup(section_name((((connection_data *)data)->p)->config_section));
	return 0;
}

/*************************************************************
* ENTRY METHOD
**************************************************************/
/*#Device.XMPP.Connection.{i}.!UCI:xmpp/connection/dmmap_cwmp_xmpp*/
static int browseXMPPConnectionInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p;
	char *connection_name = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	dmuci_get_option_value_string("xmpp", "xmpp", "conn_req_connection", &connection_name);

	synchronize_specific_config_sections_with_dmmap("xmpp", "connection", "dmmap_xmpp", &dup_list);
	list_for_each_entry(p, &dup_list, list) {
		connection_data cd = {p, "0", "0", "0", "0"};

 		if (strcmp(connection_name, section_name(p->config_section)) == 0) {
 			FILE *fp = fopen(XMPP_STATS_PATH, "r");
 			if (fp) {
 				fscanf(fp, "%31s %31s %31s %31s", cd.rx_msg, cd.tx_msg, cd.rx_err_msg, cd.tx_err_msg);
				fclose(fp);
			}
		}

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "con_inst", "con_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)&cd, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;

}

/*#Device.XMPP.Connection.{i}.!UCI:xmpp/connection_server/dmmap_cwmp_xmpp*/
static int browseXMPPConnectionServerInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct uci_section *connection_s = (((connection_data *)prev_data)->p)->config_section;
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap_eq("xmpp", "connection_server", "dmmap_xmpp", "con_name", section_name(connection_s), &dup_list);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "con_srv_inst", "con_srv_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

/*************************************************************
* ADD & DEL OBJ
**************************************************************/
static int addObjXMPPConnection(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char sec_name[32];

	snprintf(sec_name, sizeof(sec_name), "connection_%s", *instance);

	dmuci_add_section("xmpp", "connection", &s);
	dmuci_rename_section_by_section(s, sec_name);
	dmuci_set_value_by_section(s, "enable", "0");
	dmuci_set_value_by_section(s, "interval", "30");
	dmuci_set_value_by_section(s, "attempt", "16");
	dmuci_set_value_by_section(s, "serveralgorithm", "DNS-SRV");

	dmuci_add_section_bbfdm("dmmap_xmpp", "connection", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", sec_name);
	dmuci_set_value_by_section(dmmap_s, "con_inst", *instance);
	return 0;
}

static int delObjXMPPConnection(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *connection_s = (((connection_data *)data)->p)->config_section;
	struct uci_section *s = NULL, *dmmap_section = NULL, *stmp = NULL;
	
	switch (del_action) {
		case DEL_INST:
			uci_foreach_option_eq_safe("xmpp", "connection_server", "con_name", section_name(connection_s), stmp, s) {
				get_dmmap_section_of_config_section("dmmap_xmpp", "connection_server", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}

			dmuci_delete_by_section((((connection_data *)data)->p)->config_section, NULL, NULL);
			dmuci_delete_by_section((((connection_data *)data)->p)->dmmap_section, NULL, NULL);
			return 0;
		case DEL_ALL:
			uci_foreach_sections_safe("xmpp", "connection", stmp, s) {
				get_dmmap_section_of_config_section("dmmap_xmpp", "connection", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}

			uci_foreach_sections_safe("xmpp", "connection_server", stmp, s) {
				get_dmmap_section_of_config_section("dmmap_xmpp", "connection_server", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}

			return 0;
	}
	return 0;
}

static int addObjXMPPConnectionServer(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *connection_s = (((connection_data *)data)->p)->config_section;
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char sec_name[64];

	snprintf(sec_name, sizeof(sec_name), "%s_srv_%s", section_name(connection_s), *instance);

	dmuci_add_section("xmpp", "connection_server", &s);
	dmuci_rename_section_by_section(s, sec_name);
	dmuci_set_value_by_section(s, "con_name", section_name(connection_s));
	dmuci_set_value_by_section(s, "enable", "0");
	dmuci_set_value_by_section(s, "priority", "0");
	dmuci_set_value_by_section(s, "weight", "-1");
	dmuci_set_value_by_section(s, "port", "5222");

	dmuci_add_section_bbfdm("dmmap_xmpp", "connection_server", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", sec_name);
	dmuci_set_value_by_section(dmmap_s, "con_srv_inst", *instance);
	return 0;
}

static int delObjXMPPConnectionServer(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *connection_s = (((connection_data *)data)->p)->config_section;
	struct uci_section *s = NULL, *dmmap_section = NULL, *stmp = NULL;

	switch (del_action) {
		case DEL_INST:
			dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
			dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			uci_foreach_option_eq_safe("xmpp", "connection_server", "con_name", section_name(connection_s), stmp, s) {
				get_dmmap_section_of_config_section("dmmap_xmpp", "connection_server", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}
			break;
	}
	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
/*#Device.ManagementServer.ConnReqAllowedJabberIDs!UCI:xmpp/xmpp,xmpp/allowed_jid*/
static int get_management_server_conn_rep_allowed_jabber_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_option_value_string("xmpp", "xmpp", "allowed_jid", value);
	return 0;
}

static int set_management_server_conn_rep_allowed_jabber_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, 32, -1, -1, 256, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value("xmpp", "xmpp", "allowed_jid", value);
			return 0;
	}
	return 0;
}

static int get_management_server_conn_req_jabber_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *username = NULL;
	char *domain = NULL;
	char *resource = NULL;
	char *xmpp_connection = NULL;

	dmuci_get_option_value_string("xmpp", "xmpp", "conn_req_connection", &xmpp_connection);
	dmuci_get_option_value_string("xmpp", xmpp_connection, "username", &username);
	dmuci_get_option_value_string("xmpp", xmpp_connection, "domain", &domain);
	dmuci_get_option_value_string("xmpp", xmpp_connection, "resource", &resource);
	if (*username && *domain)
		dmasprintf(value, "%s@%s/%s", username, domain, resource);
	else
		*value = "";

	return 0;
}

static int get_management_server_conn_req_xmpp_connection(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *xmpp_connection = NULL;

	dmuci_get_option_value_string("xmpp", "xmpp", "conn_req_connection", &xmpp_connection);
	adm_entry_get_linker_param(ctx, "Device.XMPP.Connection.", xmpp_connection, value);
	return 0;
}

static int set_management_server_conn_req_xmpp_connection(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *allowed_objects[] = {"Device.XMPP.Connection.", NULL};
	char *linker = NULL;

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
				return FAULT_9007;

			if (dm_entry_validate_allowed_objects(ctx, value, allowed_objects))
				return FAULT_9007;

			return 0;
		case VALUESET:
			adm_entry_get_linker_value(ctx, value, &linker);
			dmuci_set_value("xmpp", "xmpp", "conn_req_connection", linker ? linker : "");
			return 0;
	}
	return 0;
}

/*#Device.XMPP.ConnectionNumberOfEntries!UCI:xmpp/connection/*/
static int get_XMPP_ConnectionNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseXMPPConnectionInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_XMPP_SupportedServerConnectAlgorithms(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "DNS-SRV,DNS,ServerTable";
	return 0;
}

/*#Device.XMPP.Connection.{i}.Enable!UCI:xmpp/connection,@i-1/enable*/
static int get_XMPPConnection_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "enable", "1");
	return 0;
}

static int set_XMPPConnection_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			return 0;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "enable", b ? "1" : "0");
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Alias!UCI:dmmap_xmpp/connection,@i-1/con_alias*/
static int get_XMPPConnection_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->dmmap_section, "con_alias", value);
	if ((*value)[0] == '\0')
		dmasprintf(value, "cpe-%s", instance);
	return 0;
}

static int set_XMPPConnection_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->dmmap_section, "con_alias", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Username!UCI:xmpp/connection,@i-1/username*/
static int get_XMPPConnection_Username(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "username", value);
	return 0;
}

static int set_XMPPConnection_Username(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 256, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "username", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Password!UCI:xmpp/connection,@i-1/password*/
static int get_XMPPConnection_Password(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "";
	return 0;
}

static int set_XMPPConnection_Password(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 256, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "password", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Domain!UCI:xmpp/connection,@i-1/domain*/
static int get_XMPPConnection_Domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "domain", value);
	return 0;
}

static int set_XMPPConnection_Domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "domain", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Resource!UCI:xmpp/connection,@i-1/resource*/
static int get_XMPPConnection_Resource(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "resource", value);
	return 0;
}

static int set_XMPPConnection_Resource(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "resource", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.ServerConnectAlgorithm!UCI:xmpp/connection,@i-1/serveralgorithm*/
static int get_XMPPConnection_ServerConnectAlgorithm(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "serveralgorithm", value);
	return 0;
}

static int set_XMPPConnection_ServerConnectAlgorithm(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, ServerConnectAlgorithm, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "serveralgorithm", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.KeepAliveInterval!UCI:xmpp/connection,@i-1/interval*/
static int get_XMPPConnection_KeepAliveInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "interval", "-1");
	return 0;
}

static int set_XMPPConnection_KeepAliveInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_long(ctx, value, RANGE_ARGS{{"-1",NULL}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "interval", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.ServerConnectAttempts!UCI:xmpp/connection,@i-1/attempt*/
static int get_XMPPConnection_ServerConnectAttempts(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "attempt", "16");
	return 0;
}

static int set_XMPPConnection_ServerConnectAttempts(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{NULL,NULL}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "attempt", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.ServerRetryInitialInterval!UCI:xmpp/connection,@i-1/initial_retry_interval*/
static int get_XMPPConnection_ServerRetryInitialInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "initial_retry_interval", "60");
	return 0;
}

static int set_XMPPConnection_ServerRetryInitialInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1","65535"}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "initial_retry_interval", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.ServerRetryIntervalMultiplier!UCI:xmpp/connection,@i-1/retry_interval_multiplier*/
static int get_XMPPConnection_ServerRetryIntervalMultiplier(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "retry_interval_multiplier", "2000");
	return 0;
}

static int set_XMPPConnection_ServerRetryIntervalMultiplier(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1000","65535"}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "retry_interval_multiplier", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.ServerRetryMaxInterval!UCI:xmpp/connection,@i-1/retry_max_interval*/
static int get_XMPPConnection_ServerRetryMaxInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "retry_max_interval", "30720");
	return 0;
}

static int set_XMPPConnection_ServerRetryMaxInterval(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"1",NULL}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "retry_max_interval", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.UseTLS!UCI:xmpp/connection,@i-1/usetls*/
static int get_XMPPConnection_UseTLS(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def((((connection_data *)data)->p)->config_section, "usetls", "1");
	return 0;
}

static int set_XMPPConnection_UseTLS(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			return 0;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section((((connection_data *)data)->p)->config_section, "usetls", b ? "1" : "0");
			return 0;
	}
	return 0;
}

static int get_XMPPConnection_JabberID(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *username = NULL;
	char *domain = NULL;
	char *resource = NULL;

	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "username", &username);
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "domain", &domain);
	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "resource", &resource);
	if (*username && *domain)
		dmasprintf(value, "%s@%s/%s", username, domain, resource);
	else
		*value = "";
	return 0;
}

/*#Device.XMPP.Connection.{i}.Status!UCI:xmpp/connection,@i-1/enable*/
static int get_XMPPConnection_Status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *status;

	dmuci_get_value_by_section_string((((connection_data *)data)->p)->config_section, "enable", &status);
	*value = (strcmp(status, "1") == 0) ? "Enabled" : "Disabled";
	return 0;
}

static int get_XMPPConnection_ServerNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseXMPPConnectionServerInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

/*#Device.XMPP.Connection.{i}.Server.{i}.Enable!UCI:xmpp/connection,@i-1/enable*/
static int get_XMPPConnectionServer_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "enable", "1");
	return 0;
}

static int set_XMPPConnectionServer_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			return 0;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "enable", b ? "1" : "0");
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Server.{i}.Alias!UCI:dmmap_xmpp/connection_server,@i-1/con_srv_alias*/
static int get_XMPPConnectionServer_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->dmmap_section, "con_srv_alias", value);
	if ((*value)[0] == '\0')
		dmasprintf(value, "cpe-%s", instance);
	return 0;
}

static int set_XMPPConnectionServer_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "con_srv_alias", value);
			return 0;
	}
	return 0;
}

static int get_XMPPConnectionServer_Priority(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "priority", value);
	return 0;
}

static int set_XMPPConnectionServer_Priority(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"0","65535"}}, 1))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "priority", value);
			break;
	}
	return 0;
}

static int get_XMPPConnectionServer_Weight(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "weight", "-1");
	return 0;
}

static int set_XMPPConnectionServer_Weight(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_long(ctx, value, RANGE_ARGS{{"-1","65535"}}, 1))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "weight", value);
			break;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Server.{i}.ServerAddress!UCI:xmpp/connection,@i-1/server_address*/
static int get_XMPPConnectionServer_ServerAddress(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "server_address", value);
	return 0;
}

static int set_XMPPConnectionServer_ServerAddress(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 256, NULL, NULL))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "server_address", value);
			return 0;
	}
	return 0;
}

/*#Device.XMPP.Connection.{i}.Server.{i}.Port!UCI:xmpp/connection,@i-1/port*/
static int get_XMPPConnectionServer_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "port", "5222");
	return 0;
}

static int set_XMPPConnectionServer_Port(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"0","65535"}}, 1))
				return FAULT_9007;
			return 0;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "port", value);
			return 0;
	}
	return 0;
}

static int get_XMPPConnectionStats_ReceivedMessages(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = data ? dmstrdup(((connection_data *)data)->rx_msg) : "";
	return 0;
}

static int get_XMPPConnectionStats_TransmittedMessages(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = data ? dmstrdup(((connection_data *)data)->tx_msg) : "";
	return 0;
}

static int get_XMPPConnectionStats_ReceivedErrorMessages(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = data ? dmstrdup(((connection_data *)data)->rx_err_msg) : "";
	return 0;
}

static int get_XMPPConnectionStats_TransmittedErrorMessages(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = data ? dmstrdup(((connection_data *)data)->tx_err_msg) : "";
	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & LEAF DEFINITION
***********************************************************************************************************************************/
/* *** Device.ManagementServer. *** */
DMLEAF tDeviceManagementServerParam[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"ConnReqAllowedJabberIDs", &DMWRITE, DMT_STRING, get_management_server_conn_rep_allowed_jabber_id, set_management_server_conn_rep_allowed_jabber_id, BBFDM_CWMP},
{"ConnReqJabberID", &DMREAD, DMT_STRING, get_management_server_conn_req_jabber_id, NULL, BBFDM_CWMP},
{"ConnReqXMPPConnection", &DMWRITE, DMT_STRING, get_management_server_conn_req_xmpp_connection, set_management_server_conn_req_xmpp_connection, BBFDM_CWMP},
{0}
};

/* *** Device.XMPP. *** */
DMOBJ tDeviceXMPPObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"XMPP", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tXMPPObj, tXMPPParams, NULL, BBFDM_BOTH},
{0}
};

DMOBJ tXMPPObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"Connection", &DMWRITE, addObjXMPPConnection, delObjXMPPConnection, NULL, browseXMPPConnectionInst, NULL, NULL, tXMPPConnectionObj, tXMPPConnectionParams, get_linker_xmpp_connection, BBFDM_BOTH, LIST_KEY{"Alias", "Username", "Domain", "Resource", NULL}},
{0}
};

DMLEAF tXMPPParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"ConnectionNumberOfEntries", &DMREAD, DMT_UNINT, get_XMPP_ConnectionNumberOfEntries, NULL, BBFDM_BOTH},
{"SupportedServerConnectAlgorithms", &DMREAD, DMT_STRING, get_XMPP_SupportedServerConnectAlgorithms, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.XMPP.Connection.{i}. *** */
DMOBJ tXMPPConnectionObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"Server", &DMWRITE, addObjXMPPConnectionServer, delObjXMPPConnectionServer, NULL, browseXMPPConnectionServerInst, NULL, NULL, NULL, tXMPPConnectionServerParams, NULL, BBFDM_BOTH, LIST_KEY{"Alias", "ServerAddress", "Port", NULL}},
{"Stats", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, NULL, tXMPPConnectionStatsParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tXMPPConnectionParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_XMPPConnection_Enable, set_XMPPConnection_Enable, BBFDM_BOTH},
{"Alias", &DMWRITE, DMT_STRING, get_XMPPConnection_Alias, set_XMPPConnection_Alias, BBFDM_BOTH},
{"Username", &DMWRITE, DMT_STRING, get_XMPPConnection_Username, set_XMPPConnection_Username, BBFDM_BOTH},
{"Password", &DMWRITE, DMT_STRING, get_XMPPConnection_Password, set_XMPPConnection_Password, BBFDM_BOTH},
{"Domain", &DMWRITE, DMT_STRING, get_XMPPConnection_Domain, set_XMPPConnection_Domain, BBFDM_BOTH},
{"Resource", &DMWRITE, DMT_STRING, get_XMPPConnection_Resource, set_XMPPConnection_Resource, BBFDM_BOTH},
{"JabberID", &DMREAD, DMT_STRING, get_XMPPConnection_JabberID, NULL, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_XMPPConnection_Status, NULL, BBFDM_BOTH},
{"ServerConnectAlgorithm", &DMWRITE, DMT_STRING, get_XMPPConnection_ServerConnectAlgorithm, set_XMPPConnection_ServerConnectAlgorithm, BBFDM_BOTH},
{"KeepAliveInterval", &DMWRITE, DMT_LONG, get_XMPPConnection_KeepAliveInterval, set_XMPPConnection_KeepAliveInterval, BBFDM_BOTH},
{"ServerConnectAttempts", &DMWRITE, DMT_UNINT, get_XMPPConnection_ServerConnectAttempts, set_XMPPConnection_ServerConnectAttempts, BBFDM_BOTH},
{"ServerRetryInitialInterval", &DMWRITE, DMT_UNINT, get_XMPPConnection_ServerRetryInitialInterval, set_XMPPConnection_ServerRetryInitialInterval, BBFDM_BOTH},
{"ServerRetryIntervalMultiplier", &DMWRITE, DMT_UNINT, get_XMPPConnection_ServerRetryIntervalMultiplier, set_XMPPConnection_ServerRetryIntervalMultiplier, BBFDM_BOTH},
{"ServerRetryMaxInterval", &DMWRITE, DMT_UNINT, get_XMPPConnection_ServerRetryMaxInterval, set_XMPPConnection_ServerRetryMaxInterval, BBFDM_BOTH},
{"UseTLS", &DMWRITE, DMT_BOOL, get_XMPPConnection_UseTLS, set_XMPPConnection_UseTLS, BBFDM_BOTH},
{"ServerNumberOfEntries", &DMREAD, DMT_UNINT, get_XMPPConnection_ServerNumberOfEntries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.XMPP.Connection.{i}.Server.{i}. *** */
DMLEAF tXMPPConnectionServerParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_XMPPConnectionServer_Enable, set_XMPPConnectionServer_Enable, BBFDM_BOTH},
{"Alias", &DMWRITE, DMT_STRING, get_XMPPConnectionServer_Alias, set_XMPPConnectionServer_Alias, BBFDM_BOTH},
{"Priority", &DMWRITE, DMT_UNINT, get_XMPPConnectionServer_Priority, set_XMPPConnectionServer_Priority, BBFDM_BOTH},
{"Weight", &DMWRITE, DMT_LONG, get_XMPPConnectionServer_Weight, set_XMPPConnectionServer_Weight, BBFDM_BOTH},
{"ServerAddress", &DMWRITE, DMT_STRING, get_XMPPConnectionServer_ServerAddress, set_XMPPConnectionServer_ServerAddress, BBFDM_BOTH},
{"Port", &DMWRITE, DMT_UNINT, get_XMPPConnectionServer_Port, set_XMPPConnectionServer_Port, BBFDM_BOTH},
{0}
};

/* *** Device.XMPP.Connection.{i}.Stats. *** */
DMLEAF tXMPPConnectionStatsParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"ReceivedMessages", &DMREAD, DMT_UNINT, get_XMPPConnectionStats_ReceivedMessages, NULL, BBFDM_BOTH},
{"TransmittedMessages", &DMREAD, DMT_UNINT, get_XMPPConnectionStats_TransmittedMessages, NULL, BBFDM_BOTH},
{"ReceivedErrorMessages", &DMREAD, DMT_UNINT, get_XMPPConnectionStats_ReceivedErrorMessages, NULL, BBFDM_BOTH},
{"TransmittedErrorMessages", &DMREAD, DMT_UNINT, get_XMPPConnectionStats_TransmittedErrorMessages, NULL, BBFDM_BOTH},
{0}
};
